'use strict';
const util = require('util');
const {spawn} = require('child_process');
const path = require('path');
const {format} = require('util');
const execa = require('execa');
const fse = require('fs-extra');
const importLazy = require('import-lazy')(require);
const _ = require('lodash');



const configstore = importLazy('configstore');
const chalk = importLazy('chalk');
const semverDiff = importLazy('semver-diff');
const latestVersion = importLazy('latest-version');
const isNpm = importLazy('is-npm');
const isInstalledGlobally = importLazy('is-installed-globally');
const isYarnGlobal = importLazy('is-yarn-global');
const hasYarn = importLazy('has-yarn');
const boxen = importLazy('boxen');
const xdgBasedir = importLazy('xdg-basedir');
const isCi = importLazy('is-ci');

const mapWodaXConfigs = require('./wodax-config');

/**
 * 判断是否为当前包
 * @param packageName 包名
 * @returns {boolean}
 */
function matchCurrentPackage(packageName) {
	try {
		const curPKG = require(`./package`);
		return packageName === curPKG.name;
	}catch (e) {
		console.error(e)
	}
	return false;
}

const ONE_DAY = 1000 * 60 * 60 * 24;

class UpdateNotifier {
	constructor(options = {}) {
		this.options = options;
		options.pkg = options.pkg || {};
		options.distTag = options.distTag || 'latest';

		// Reduce pkg to the essential keys. with fallback to deprecated options
		options.pkg = {
			name: options.pkg.name || options.packageName,
			version: options.pkg.version || options.packageVersion,
			dependenciesType: options.pkg.dependenciesType || options.dependenciesType
		};

		if (!options.pkg.name || !options.pkg.version) {
			throw new Error('pkg.name and pkg.version required');
		}

		options.runCheckByFuncName = options.runCheckByFuncName || 'checkNpm'; /// default use checkNpm
		options.templateClassId = options.templateClassId || ''; /// eg. "00-wodax-ebmp"
		options.templateKey = options.templateKey || ''; /// eg. "business-app-backstage"
		options.templateVersion = options.templateVersion || ''; /// eg. "0.1.0"

		this.packageName = options.pkg.name;
		this.packageVersion = options.pkg.version;
		this.dependenciesType = options.pkg.dependenciesType || 'dependencies';

		this.templateClassId = options.templateClassId;
		this.templateKey = options.templateKey;
		this.templateVersion = options.templateVersion;

		this.runCheckByFuncName = options.runCheckByFuncName;

		this.updateCheckInterval = typeof options.updateCheckInterval === 'number' ? options.updateCheckInterval : ONE_DAY;
		this.hasCallback = typeof options.callback === 'function';
		this.callback = options.callback || (() => {
		});
		this.disabled = 'NO_UPDATE_NOTIFIER' in process.env ||
			process.argv.includes('--no-update-notifier') ||
			isCi();
		this.shouldNotifyInNpmScript = options.shouldNotifyInNpmScript;

		if (!this.disabled) {
			try {
				const ConfigStore = configstore();
				this.config = new ConfigStore(`update-notifier-${this.packageName}`, {
					optOut: false,
					// Init with the current time so the first check is only
					// after the set interval, so not to bother users right away
					lastUpdateCheck: Date.now()
				});
			} catch (error) {
				// Expecting error code EACCES or EPERM
				const message =
					chalk().yellow(format(' %s update check failed ', options.pkg.name)) +
					format('\n Try running with %s or get access ', chalk().cyan('sudo')) +
					'\n to the local update config store via \n' +
					chalk().cyan(format(' sudo chown -R $USER:$(id -gn $USER) %s ', xdgBasedir().config));

				process.on('exit', () => {
					console.error('\n' + boxen()(message, {align: 'center'}));
				});
			}
		}
	}

	/**
	 * Check the update informations
	 */
	check() {
		// Check for Wodax-project Or Npm
		const wodaxConfig = this.getWodaXConfig(this.templateClassId, this.templateKey);
		this.options.runCheckByFuncName = wodaxConfig ? 'checkWodaX' : 'checkNpm';

		if (this.hasCallback) {
			(async () => {
				try {
					const fn = wodaxConfig ? this.checkWodaX : this.checkNpm;
					this.callback(null, await fn.bind(this)(), this);
				} catch (error) {
					this.callback(error);
				}
			})();

			return;
		}

		if (
			!this.config ||
			this.config.get('optOut') ||
			this.disabled
		) {
			return;
		}

		this.update = this.config.get('update');

		if (this.update) {
			this.config.delete('update');
		}

		// Only check for updates on a set interval
		if (Date.now() - this.config.get('lastUpdateCheck') < this.updateCheckInterval) {
			return;
		}

		// Spawn a detached process, passing the options as an environment property
		const argvString = JSON.stringify(this.options);
		console.log(argvString);
		spawn(process.execPath, [path.join(__dirname, 'check.js'), argvString], {
			detached: true,
			stdio: 'ignore',
			env: {
				...process.env,
				'NODE_OPTIONS':"--max_old_space_size=4096"
			}
		}).unref();
	}

	/** Get config information from wodax-config.js
	 * @param {String} templateClassId  eg："00-wodax-ebmp"
	 * @param {String} templateKey  eg："business-app-backstage"
	 * @returns {null} config object, maybe null or undefended
	 */
	getWodaXConfig(templateClassId, templateKey) {
		if (!this.templateClassId || !this.templateKey) {
			return null;
		}

		return _.get(mapWodaXConfigs, [templateClassId, templateKey]) || null;
	}

	checkInSemverList(versionType) {
		if (!!!versionType) return false;
		return [
			`patch`,
			`prepatch`,
			`major`,
			`premajor`,
			`minor`,
			`preminor`
		].includes(versionType)
	}

	printSemverInfo({packageName, packageVersion, newVersion, versionType}) {
		if (this.checkInSemverList(versionType)) {
			console.log(`${chalk().gray(`>>> ◢▆▅▄▃▃▄▅▇◣`)} 检测到: >>> [${packageName}@${packageVersion} -> ${newVersion}] ....`);
		}
	}

	async checkWodaX() {
		this.options.runCheckByFuncName = 'checkWodaX';
		const config = this.getWodaXConfig(this.templateClassId, this.templateKey) || {};
		const keyStable = 'stable';
		const keyDiscard = 'discard';
		const keyTemplateVersion = 'version';
		const repeatNum = 3;



		const tplVersion = _.get(config, [keyTemplateVersion]);
		if (tplVersion && this.templateVersion) {
			// const verType = semverDiff()(this.templateVersion, tplVersion);
			// if (verType) {
			//
			// }
			const projectName = `"${this.templateClassId}">"${this.templateKey}"`;
			const gitUrl = _.get(config, ['vcs', 'url']);
			const gitBranch = _.get(config, ['vcs', 'branch']);

			const information = `` + '\n' +
				` 1.[模板名称] ${chalk().blue.underline(projectName)}` + '\n' +
				` 2.[版本信息] 当前：${chalk().cyan.underline(this.templateVersion)} ${chalk().reset(' → ')} 远端：${chalk().cyan.underline(tplVersion)}` + '\n' +
				` 3.[仓库地址] ${gitUrl} ` + '\n' +
				` 4.[仓库分支] ${gitBranch} ` + '\n' +
				` ${chalk().red('-'.repeat(60))}`;


			console.log(`${chalk().red('▶ '.repeat(repeatNum))} 模板说明：\n` +
				`${information}`);
		}


		const pkgDiscard = _.get(config, [keyDiscard, this.packageName]);
		if (pkgDiscard) {
			const projectName = `${chalk().red.underline.bold(this.packageName)}`;
			const information = `${chalk().red.bold('已经被废弃，不建议使用 ... ✘ ')}`;
			console.log(`${chalk().black('☠ '.repeat(repeatNum))} ${projectName} : ${information}`);
		}

		const {distTag} = this.options;
		const pkgStableVersion = _.get(config, [keyStable, this.packageName]);
		if (pkgStableVersion) {
			let versionType = distTag;
			try {
				// console.log(`call checkWodaX`, this.packageName, this.packageVersion, pkgStableVersion);
				const fmtPackageVersion = this.packageVersion.replace(/\^/g,``);
				versionType = semverDiff()(fmtPackageVersion, pkgStableVersion) || distTag;
				// console.log(`${this.packageName} - ${versionType}`);
			} catch(e){
				console.error(e);
			}

			this.printSemverInfo({
				packageName: this.packageName,
				packageVersion: this.packageVersion,
				newVersion: pkgStableVersion,
				versionType: versionType
			});

			return {
				latest: pkgStableVersion,
				current: this.packageVersion,
				type: versionType,
				name: this.packageName,
				scope: 'wodax'
			};
		}

		const info = await this.checkNpm();
		return info;
	}

	async checkNpm() {
		const {distTag} = this.options;
		const latest = await latestVersion()(this.packageName, {version: distTag});
		this.options.runCheckByFuncName = 'checkNpm';

		let versionType = null;
		try {
			const fmtPackageVersion = this.packageVersion.replace(/\^/g,``);
			if (fmtPackageVersion !== 'latest') {
				versionType = semverDiff()(fmtPackageVersion, latest);
			}
		} catch(e){
			console.error(e);
		}

		this.printSemverInfo({
			packageName: this.packageName,
			packageVersion: this.packageVersion,
			newVersion: latest,
			versionType: versionType
		});

		return {
			latest,
			current: this.packageVersion,
			type: versionType,
			name: this.packageName,
			scope: 'npm'
		};
	}

	/** Notify
	 * @param {Object} options  a init options
	 * @returns {UpdateNotifier} a UpdateNotifier Object
	 */
	notify(options) {
		const suppressForNpm = !this.shouldNotifyInNpmScript && isNpm().isNpm;
		if (suppressForNpm || !this.update) {
			return this;
		}

		options = {
			isGlobal: isInstalledGlobally(),
			isYarnGlobal: isYarnGlobal()(),
			...options
		};

		const installCommand = this.getInstallCommand(options);

		const forWodaX = ((this.options.runCheckByFuncName === 'checkWodaX') || matchCurrentPackage(this.packageName));
		const messageCaption = forWodaX ?
			// chalk().black.bgWhite.underline.bold(' WodaX 项目依赖强制检测 ') + '\n' :
			chalk().underline.bold(' WodaX 项目依赖强制检测 ') + '\n' :
			'';
		const captionUpdate = forWodaX ? '可用更新:': '发现可用更新';
		const tipUpdate = forWodaX ? '\n使用 ': '\n【非强制核心更新】 要更新，可使用 ';

		options.message = options.message || messageCaption + `${captionUpdate}： ${this.packageName} ` + chalk().dim(this.update.current) + chalk().dim(' → ') +
			chalk().green(this.update.latest) + `${tipUpdate}` + chalk().cyan(installCommand) + ' 命令更新依赖';

		options.boxenOpts = options.boxenOpts || {
			padding: 1,
			margin: 1,
			align: 'center',
			backgroundColor: options.isMilestone ? '#000000' : '',
			borderColor: options.isMilestone ? '#ff0000' : 'yellow',
			borderStyle: options.isMilestone ? 'bold' : 'round'
		};


		const message = '\n' + boxen()(options.message, options.boxenOpts);

		if (options.defer === false) {
			console.error(message);
		} else {
			process.on('exit', () => {
				console.error(message);
			});

			process.on('SIGINT', () => {
				console.error('');
				process.exit();
			});
		}

		return this;
	}

	printInfo(options) {
		options = {
			isGlobal: isInstalledGlobally(),
			isYarnGlobal: isYarnGlobal()(),
			...options
		};

		this.update = this.config.get('update');

		const installCommand = this.getInstallCommand({options, latestVersion: this.update.latest});

		const forWodaX = ((this.options.runCheckByFuncName === 'checkWodaX') || matchCurrentPackage(this.packageName));
		const messageCaption = forWodaX ?
			// chalk().black.bgWhite.underline.bold(' WodaX 项目依赖强制检测 ') + '\n' :
			chalk().underline.bold(' WodaX 项目依赖强制检测 ') + '\n' :
			'';
		const captionUpdate = forWodaX ? '可用更新:': '发现可用更新';
		const tipUpdate = forWodaX ? '\n使用 ': '\n【非强制核心更新】 要更新，可使用 ';

		options.message = options.message || messageCaption + `${captionUpdate}： ${this.packageName} ` + chalk().dim(this.update.current) + chalk().dim(' → ') +
			chalk().green(this.update.latest) + `${tipUpdate}` + chalk().cyan(installCommand) + ' 命令更新依赖';

		options.isMilestone = options.isMilestone || false;
		options.boxenOpts = options.boxenOpts || {
			padding: options.isMilestone ? 2 : 1,
			margin: options.isMilestone ? 2 : 1,
			align: 'center',
			backgroundColor: options.isMilestone ? '#000000' : '',
			borderColor: options.isMilestone ? '#ff0000' : 'yellow',
			borderStyle: options.isMilestone ? 'bold' : 'round'
		};

		const message = '\n' + boxen()(options.message, options.boxenOpts);
		console.log(message);

		// Note: Try update the wodax project
		try {
			if (forWodaX && !matchCurrentPackage(this.packageName)) {
				console.log(`强制依赖更新：正在尝试执行 ${chalk().cyan(installCommand)} 命令，请耐心等待 ...`);
				console.log(`如果更新失败，请在终端中手动运行该命令 '${installCommand}' ...`);
				execa.commandSync(installCommand, {
					timeout: 1000 * 60 * 10, /// 单位毫秒,
					env: {
						...process.env,
						'NODE_OPTIONS':"--max_old_space_size=4096"
					}
				});
			}

		} catch (e) {
			console.error(e);
			process.exit(1);
		}

		return this;
	}

	static GetCurrentRunIsNPMCLI() {
		return /npm-cli\.js$|npm-cli$/.test(process.env['npm_execpath'])
	}
	static GetCurrentRunIsYarnCLI() {
		return /yarn\.js$|yarnpkg$/.test(process.env['npm_execpath'])
	}
	static GetCurrentRunIsCNPMCLI(){
		return /cnpm$/.test(process.env['npm_execpath'])
	}
	static GetCurrentRunISPNMPCLI(){
		return /pnpm\.js$/.test(process.env['npm_execpath'])
	}
	static GetHasYarnLockFile(){
		let exist = false;
		try {
			const dir = `${path.normalize(path.join(__dirname, `../../`, `yarn.lock`))}`;
			exist = fse.pathExistsSync(`${dir}`);
		}catch(e){}
		return exist;
	}


	static GetInstallCommand(options) {
		let installCommand;
		const dependenciesOptions = `${{
			devDependencies: '-D',
			optionalDependencies: '-O',
			dependencies: '',
			peerDependency: '-P',
			bundleDependencies: '-B'
		}[options.dependenciesType]}`;

		const yarnCommandGlobal = `yarn global add ${options.packageName}@${options.latestVersion}`;
		const yarnCommand = `yarn add ${options.packageName}@${options.latestVersion} ${dependenciesOptions}`;
		const npmCommand = `npm i ${options.isGlobal ? '-g ' : ''}${options.packageName}@${options.latestVersion} ${dependenciesOptions}`;

		/// 强化使用yarn来安装包
		if (options.isYarnGlobal) {
			installCommand = yarnCommandGlobal;
		} else if (UpdateNotifier.GetCurrentRunIsYarnCLI()) {
			installCommand = yarnCommand;
		} else if (hasYarn()() && UpdateNotifier.GetHasYarnLockFile()) {
			installCommand = yarnCommand;
		} else if (UpdateNotifier.GetCurrentRunIsNPMCLI()){
			installCommand = npmCommand;
		} else if (hasYarn()()){
			installCommand = yarnCommand;
		} else {
			installCommand = npmCommand;
		}

		return installCommand;
	}

	getInstallCommand(options) {
		let installCommand = UpdateNotifier.GetInstallCommand({
			dependenciesType: this.dependenciesType,
			packageName: this.packageName,
			latestVersion: options.latestVersion,
			isYarnGlobal: options.isYarnGlobal,
			isGlobal: options.isGlobal,
		});

		return installCommand;
	}
}

const createUpdateNotifier = options => {
	const { pkg } = {pkg:{name: ''}, ...options};
	const { name, version } = {name:'', version:'', ...pkg};
	console.log(`${chalk().gray(`>>> ◢▆▅▄▃▃▄▅▇◣`)} 正在执行依赖更新检测程序: >>> [${name}@${version}] ....`);
	const updateNotifier = new UpdateNotifier(options);
	updateNotifier.check();
	return updateNotifier;
};

function CheckGitSettings() {
	[
		`git config core.autocrlf false`,   // 让Git不要管Windows/Unix 换行符转换的事情
		`git config core.ignorecase false`, // 默认开启大小写感知
		`git config core.quotepath off`,    // 避免git status显示的中文文件名乱码
		`git config gui.encoding utf-8`,    // 避免git gui中的中文乱码
	].forEach((command) => {
		try {
			require('child_process').execSync(command)
		} catch (e) {
		}
	});
}

function SelfCheck(cb) {
	const pkgInfo = require('./package.json');
	const {name: packageName, version: packageVersion} = pkgInfo;

	console.log(`${chalk().gray(`>>> ʅ（´◔౪◔）ʃ`)} 执行 '${packageName}' 自检程序 ....`);
	// 1. 检查Git环境设置
	CheckGitSettings();

	// 2. 更新模板补丁核心文件
	try {
		const patch = require('./patch');
		patch.tryApplyPatches(false)
	}catch (e) {
		console.trace(e);
	}

	// 3. 自检开始
	const dependenciesType = 'devDependencies';
	createUpdateNotifier({
		pkg: {
			name: packageName,
			version: packageVersion,
			dependenciesType: dependenciesType
		},
		updateCheckInterval: 0,
		callback(err, info, instance) {
			let needUpdate = false;
			let installCommand = UpdateNotifier.GetInstallCommand({
				dependenciesType: dependenciesType,
				packageName: packageName,
				latestVersion: `latest`,
				isGlobal: isInstalledGlobally(),
				isYarnGlobal: isYarnGlobal()(),
			});

			if (err) {
				console.error(err);
			} else {
				try {
					const { type } = info;
					console.log(`${chalk().green(`>>> ◢▆▅▄▃▃▄▅▇◣`)} 解析信息: >>> [${packageName}@${packageVersion}] [type = ${type}]....`);
					const handler = (instance !== undefined && instance !== null) ? instance : this;
					if (type && type !== 'latest') {
						needUpdate = true;
						try {
							handler.config.set('update', info);
						} catch (e) {}
						try {
							handler.printInfo({info, isMilestone: true});
						} catch (e) {}
					}
				} catch (e) {
					console.error(e);
				}
			}

			if (typeof cb === 'function') {
				cb(err, {packageName, needUpdate, installCommand});
			}
		}
	});
}

module.exports = createUpdateNotifier;
module.exports.UpdateNotifier = UpdateNotifier;
module.exports.selfCheckUpdate = SelfCheck;
