'use strict';
const glob = require('glob');
const path = require('path');
const stylelint = require('stylelint');
const styleConfigFile = require.resolve('../.stylelintrc.json');

const styles = {
  bold: ['\x1B[1m%s\x1B[22m'],
  italic: ['\x1B[3m%s\x1B[23m'],
  underline: ['\x1B[4m%s\x1B[24m'],
  inverse: ['\x1B[7m%s\x1B[27m'],
  strikethrough: ['\x1B[9m%s\x1B[29m'],
  white: ['\x1B[37m%s\x1B[39m'],
  grey: ['\x1B[90m%s\x1B[39m'],
  black: ['\x1B[30m%s\x1B[39m'],
  blue: ['\x1B[34m%s\x1B[39m'],
  cyan: ['\x1B[36m%s\x1B[39m'],
  green: ['\x1B[32m%s\x1B[39m'],
  magenta: ['\x1B[35m%s\x1B[39m'],
  red: ['\x1B[31m%s\x1B[39m'],
  yellow: ['\x1B[33m%s\x1B[39m'],
  whiteBG: ['\x1B[47m%s\x1B[49m'],
  greyBG: ['\x1B[49;5;8m%s\x1B[49m'],
  blackBG: ['\x1B[40m%s\x1B[49m'],
  blueBG: ['\x1B[44m%s\x1B[49m'],
  cyanBG: ['\x1B[46m%s\x1B[49m'],
  greenBG: ['\x1B[42m%s\x1B[49m'],
  magentaBG: ['\x1B[45m%s\x1B[49m'],
  redBG: ['\x1B[41m%s\x1B[49m'],
  yellowBG: ['\x1B[43m%s\x1B[49m'],
};

const logTaskPre = `>>> ◢▆▅▄▃▃▄▅▇◣`;

function formatContent(styleList = [], content) {
  styleList.forEach((oneStyle) => {
    content = oneStyle[0].replace(`%s`, content);
  });
  return content;
}

const totalInfo = {
  errors: [],
  warnings: [],
};

const allProcessConfig = [
  {files: 'src/**/*.css', syntax: '', fix: true},
  {files: 'src/**/*.less', syntax: 'less', fix: true},
  {files: 'src/**/*.scss', syntax: 'scss', fix: true},
  {files: 'src/**/*.sass', syntax: 'sass', fix: true},
  {files: 'src/**/*.html', syntax: 'html', fix: true},
  {files: 'src/**/*.htm', syntax: 'html', fix: true},
];
let curProcessedCount = 0;

function __print_total_info(_totalInfo) {
  // 输出总的统计信息
  const hasError = _totalInfo.errors.length > 0 ? 1 : 0;
  const hasWarning = _totalInfo.warnings.length > 0 ? 1 : 0;
  const styleConfig = {
    [`10`]: styles.magenta,
    [`11`]: styles.red,
    [`01`]: styles.yellow,
    [`00`]: styles.white,
  };
  const allErrorsWarningCount =
    _totalInfo.errors.length + _totalInfo.warnings.length;
  const info = `\n${allErrorsWarningCount} problems (${_totalInfo.errors.length} errors, ${_totalInfo.warnings.length} warnings)\n`;
  console.log(formatContent([styleConfig[`${hasError}${hasWarning}`]], info));
  return hasError;
}

function printStatisticsInfo(_totalInfo) {
  curProcessedCount += 1;
  if (curProcessedCount < allProcessConfig.length) return;
  const hasError = __print_total_info(_totalInfo);
  if (hasError) {
    process.exit(1);
  }
  process.exit(0);
}


console.log(
  [
    formatContent([styles.grey], logTaskPre),
    formatContent([styles.grey], `执行样式文件代码规范检查`),
    ``,
  ].join(` `.repeat(2))
);

allProcessConfig.forEach((item) => {
  const foundFiles = glob.sync(item.files, {
    cwd: path.join(__dirname, '../'),
    ignore: ['**/node_modules/**', 'build/**', 'dist/**'],
  });

  //console.log(foundFiles);
  if (foundFiles.length === 0) {
    printStatisticsInfo(totalInfo);
    return false;
  }

  try {
    stylelint
      .lint({
        configFile: styleConfigFile,
        files: foundFiles,
        syntax: item.syntax,
        fix: item.fix,
      })
      .then((data) => {
        try {
          if (data.errored) {
            data.results.forEach((res) => {
              const {source, warnings, errored} = res;
              if (errored) {
                console.log(
                  formatContent([styles.underline, styles.cyan], source)
                );
                warnings.forEach((oneWarning) => {
                  // line: 9,
                  //   column: 1,
                  //   rule: 'selector-pseudo-class-no-unknown',
                  //   severity: 'error',
                  //   text:
                  // 'Unexpected unknown pseudo-class selector ":global2" (selector-pseudo-class-no-unknown)' }
                  const {line, column, rule, severity, text} = oneWarning;
                  const contents = [
                    ``,
                    formatContent([styles.green], `${line}:${column}`),
                    `${(severity === 'error'
                        ? styles.red[0]
                        : styles.yellow[0]
                    ).replace(`%s`, severity)}`,
                    `${text.replace(`(${rule})`, '')}`,
                    `${styles.grey[0].replace(`%s`, rule)}`,
                  ];
                  console.log(contents.join(` `.repeat(2)));

                  (severity === 'error'
                      ? totalInfo.errors
                      : totalInfo.warnings
                  ).push(oneWarning);
                });
              }
            });
          }
        } catch (e) {
        }

        printStatisticsInfo(totalInfo);
      })
      .catch((err) => {
        console.error(err);
      });
  } catch (e) {
    console.error(e);
  }
});
