/**
 * copy to https://github.com/facebook/react/blob/master/scripts/prettier/index.js
 * prettier api doc https://prettier.io/docs/en/api.html
 *----------*****--------------
 * 代码独立，要求可以独立运行
 *  prettier all js and all ts.
 *----------*****--------------
 */

const glob = require('glob');
const prettier = require('prettier');
const fs = require('fs');
const prettierConfigPath = require.resolve('../.prettierrc');

const styles = {
  bold: ['\x1B[1m%s\x1B[22m'],
  italic: ['\x1B[3m%s\x1B[23m'],
  underline: ['\x1B[4m%s\x1B[24m'],
  inverse: ['\x1B[7m%s\x1B[27m'],
  strikethrough: ['\x1B[9m%s\x1B[29m'],
  white: ['\x1B[37m%s\x1B[39m'],
  grey: ['\x1B[90m%s\x1B[39m'],
  black: ['\x1B[30m%s\x1B[39m'],
  blue: ['\x1B[34m%s\x1B[39m'],
  cyan: ['\x1B[36m%s\x1B[39m'],
  green: ['\x1B[32m%s\x1B[39m'],
  magenta: ['\x1B[35m%s\x1B[39m'],
  red: ['\x1B[31m%s\x1B[39m'],
  yellow: ['\x1B[33m%s\x1B[39m'],
  whiteBG: ['\x1B[47m%s\x1B[49m'],
  greyBG: ['\x1B[49;5;8m%s\x1B[49m'],
  blackBG: ['\x1B[40m%s\x1B[49m'],
  blueBG: ['\x1B[44m%s\x1B[49m'],
  cyanBG: ['\x1B[46m%s\x1B[49m'],
  greenBG: ['\x1B[42m%s\x1B[49m'],
  magentaBG: ['\x1B[45m%s\x1B[49m'],
  redBG: ['\x1B[41m%s\x1B[49m'],
  yellowBG: ['\x1B[43m%s\x1B[49m'],
};

function formatContent(styleList = [], content) {
  styleList.forEach((oneStyle) => {
    content = oneStyle[0].replace(`%s`, content);
  });
  return content;
}

const logTaskPre = `>>> ◢▆▅▄▃▃▄▅▇◣`;

function __print_total_info(_totalInfo) {
  // 输出总的统计信息
  const hasError = _totalInfo.errors.length > 0 ? 1 : 0;
  const hasWarning = _totalInfo.warnings.length > 0 ? 1 : 0;
  const styleConfig = {
    [`10`]: styles.magenta,
    [`11`]: styles.red,
    [`01`]: styles.yellow,
    [`00`]: styles.white,
  };
  const allErrorsWarningCount =
    _totalInfo.errors.length + _totalInfo.warnings.length;
  const info = `${allErrorsWarningCount} problems (${_totalInfo.errors.length} errors, ${_totalInfo.warnings.length} warnings)\n`;
  console.log(formatContent([styleConfig[`${hasError}${hasWarning}`]], info));
  return hasError;
}

const globalIgnores = ['**/node_modules/**', 'scripts/**', 'build/**', 'dist/**', '**/.umi/**'];

/**
 * 针对目录及文件名的规范检查：
 * 注:项目中核心文件名称，均采用小写、下划线、$、数字组成
 */
function runCheckListForPathNameRule() {
  console.log(
    [
      formatContent([styles.grey], logTaskPre),
      formatContent([styles.grey], `执行目录名及文件名代码规范检查`),
      ``,
    ].join(` `.repeat(2))
  );

  const totalInfo = {
    errors: [],
    warnings: [],
  };

  const allFiles = [].concat(
    glob.sync('+(src|config|mock|e2e)/**', {
      ignore: globalIgnores,
    })
  );
  if (!allFiles.length) {
    return;
  }

  allFiles.forEach((file = '') => {
    try {
      const foundUpperChar = /[A-Z]/.test(file);
      if (foundUpperChar) {
        totalInfo.warnings.push(`${file}`);
        const contents = [
          formatContent([styles.yellow], `[!]`),
          formatContent([styles.green], `==>`),
          formatContent([], `存在大写字母`),
          formatContent([styles.red], `${file}`),
        ];
        console.log(contents.join(` `.repeat(2)));
      }
    } catch (e) {
      console.error(e);
    }
  });
  __print_total_info(totalInfo);
}

/**
 * 执行格式化
 */
function runFormat() {
  console.log(
    [
      formatContent([styles.grey], logTaskPre),
      formatContent([styles.grey], `执行格式化代码程序`),
      ``,
    ].join(` `.repeat(2))
  );

  let didError = false;

  let files = [];

  const jsFiles = glob.sync('**/*.js*', {
    ignore: globalIgnores,
  });
  const tsFiles = glob.sync('**/*.ts*', {
    ignore: globalIgnores,
  });
  files = files.concat(jsFiles);
  files = files.concat(tsFiles);
  if (!files.length) {
    return;
  }

  files.forEach((file) => {
    // console.log(file);
    const options = prettier.resolveConfig.sync(file, {
      config: prettierConfigPath,
    });
    const fileInfo = prettier.getFileInfo.sync(file);
    if (fileInfo.ignored) {
      return;
    }
    try {
      const input = fs.readFileSync(file, 'utf8');
      const withParserOptions = {
        ...options,
        parser: fileInfo.inferredParser,
      };
      const output = prettier.format(input, withParserOptions);
      if (output !== input) {
        fs.writeFileSync(file, output, 'utf8');
        const contents = [
          formatContent([styles.green], `[√]`),
          formatContent([styles.blue], `==> 执行格式化`),
          formatContent([styles.cyan], `${file}`),
        ];
        console.log(contents.join(` `.repeat(2)));
      }
    } catch (e) {
      didError = true;
      console.error(e);
    }
  });

  if (didError) {
    process.exit(1);
  }

  console.log(
    formatContent([styles.yellow], `prettier success (格式化代码成功)!`)
  );
}

// -> main run
runCheckListForPathNameRule();
runFormat();
